/*********************************************************************
 *
 * Authors: Vincenzo Ciaschini - Vincenzo.Ciaschini@cnaf.infn.it 
 *
 * Copyright (c) 2002, 2003 INFN-CNAF on behalf of the EU DataGrid.
 * For license conditions see LICENSE file or
 * http://www.edg.org/license.html
 *
 * Parts of this code may be based upon or even include verbatim pieces,
 * originally written by other people, in which case the original header
 * follows.
 *
 *********************************************************************/
/***************************************************************************
 *  filename  : GSISocketClient.cpp
 *  authors   : Salvatore Monforte <salvatore.monforte@ct.infn.it>
 *  copyright : (C) 2001 by INFN
 ***************************************************************************/

// $Id:

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>

/** This class header file. */
#include "Client.h"
/** The tokens transission and reception features definitions. */
#include "tokens.h"

#include "newca.h"

#include "openssl/buffer.h"
#include "openssl/crypto.h"
#include "openssl/objects.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

/**
 * Constructor.
 * @param p the secure server port.
 * @param b the backlog, that is the maximum number of outstanding connection requests.
 */
GSISocketClient::GSISocketClient(const std::string h, int p, int v) :
  version(v), host(h), port(p), context(GSS_C_NO_CONTEXT),
  credential(GSS_C_NO_CREDENTIAL), _server_contact(""), conflags(0),
  gsi_logfile(NULL), own_subject(""), own_ca(""), own_key(NULL),
  peer_subject(""), peer_ca(""), peer_key(NULL), peer_cert(NULL), opened(false)
{
  OBJ_create("0.9.2342.19200300.100.1.1","USERID","userId");
}

/**
 * Destructor.
 */  
GSISocketClient::~GSISocketClient()
{
  Close();
}

void 
GSISocketClient::SetFlags(OM_uint32 f)
{
  conflags = f;
}

static char *
get_globusid(gss_cred_id_t handle)
{
  char *globusid;
  char *globusid_tmp;
  gss_name_t server_name = GSS_C_NO_NAME;
  gss_buffer_desc server_buffer_desc = GSS_C_EMPTY_BUFFER;
  gss_buffer_t server_buffer = &server_buffer_desc;
  OM_uint32 major_status = 0;
  OM_uint32 minor_status = 0;
  OM_uint32 minor_status2 = 0;

  if ((major_status = gss_inquire_cred(&minor_status,
				       handle,
				       &server_name,
				       NULL, NULL, NULL)) == GSS_S_COMPLETE) {
    major_status = gss_display_name(&minor_status,
				    server_name, server_buffer, NULL);
    gss_release_name(&minor_status2, &server_name);
  }
  /*
   * The gssapi_cleartext does not implement gss_inquire_cred,
   * so fall back to using environment variable.
   */
  if (major_status == GSS_S_COMPLETE) {
    globusid = (char *)server_buffer_desc.value;
  }
  else {
    return NULL;
  }
  globusid_tmp = strdup(globusid);

  if (server_buffer_desc.value) {
    gss_release_buffer(&minor_status2, server_buffer);
  }
  return globusid_tmp;
}

/**
 * Initialize GSI Authentication.
 * This method asks the server for authentication.
 * @param sock the socket descriptot
 * @return true on success, false otherwise.
 */
bool 
GSISocketClient::InitGSIAuthentication(int sock)
{
   OM_uint32                   major_status = 0;
   OM_uint32                   minor_status = 0;
   OM_uint32                   status       = 0;

   OM_uint32                   req_flags  = conflags;
   OM_uint32                   ret_flags  = 0;
   int                         token_status = 0;
   gss_name_t                  targ_name;
   gss_buffer_desc             name_buffer;
   char                        service[1024];


   if (credential != GSS_C_NO_CREDENTIAL)
     gss_release_cred(&status, &credential);
   credential = GSS_C_NO_CREDENTIAL;

   if (context != GSS_C_NO_CONTEXT)
     gss_delete_sec_context(&status, &context, GSS_C_NO_BUFFER);
   context= GSS_C_NO_CONTEXT;

   /* acquire our credentials */
   major_status = globus_gss_assist_acquire_cred(&minor_status,
						 GSS_C_BOTH,
						 &credential);

   if(major_status != GSS_S_COMPLETE) {
     if (gsi_logfile)
       globus_gss_assist_display_status(gsi_logfile,
					"Failed to acquire credentials: ",
					major_status, minor_status, 0);
     return false;
   }
   
   char *tmp;

   tmp = getMCA(credential, version);
   if (tmp)
     own_ca = std::string(tmp);

   tmp = get_globusid(credential);
   if (tmp)
     own_subject = std::string(tmp);

   own_key = get_private_key(credential, version);

   snprintf(service, sizeof(service), "host@%s", host.c_str()); /* XXX */

   /* initialize the security context */
   /* credential has to be fill in beforehand */
   major_status =
     globus_gss_assist_init_sec_context(&minor_status, credential,
					&context,
					_server_contact.empty() ? service :(char*) _server_contact.c_str(), 
					req_flags, &ret_flags,
					&token_status,
					get_token, (void *) &sock,
					send_token, (void *) &sock);

   peer_subject = _server_contact.empty() ? service : _server_contact; 

   tmp = get_peer_CA(context, version);
   if (tmp)
     peer_ca = std::string(tmp);

   peer_key = get_delegated_public_key(context, version);
   peer_cert = get_delegated_cert(context, version);

   if(major_status != GSS_S_COMPLETE) {
     if (gsi_logfile)
       globus_gss_assist_display_status(gsi_logfile,
					"Failed to establish security context (init): ",
					major_status, minor_status, token_status);
     return false;
   }

   if ((ret_flags & req_flags) != req_flags) {
     if (gsi_logfile)
       fprintf(gsi_logfile, "Flags Mismatch:\nExpected: %d\nReceived:%d\n",
	       req_flags, (ret_flags & req_flags));
     return false;
   }

   return true;
}

/**
 * Open the connection.
 * @return true for successful opening, false otherwise.
 */
bool 
GSISocketClient::Open()
{
  peeraddr_in.sin_family = AF_INET;
    
  struct hostent *hp; 
    
  if (!(hp = gethostbyname(host.c_str())))
    return false;

  peeraddr_in.sin_addr.s_addr = ((struct in_addr *)(hp->h_addr))->s_addr;
  peeraddr_in.sin_port = htons(port);
  context = GSS_C_NO_CONTEXT;
  credential = GSS_C_NO_CREDENTIAL;

  if ((sck = socket(AF_INET, SOCK_STREAM, 0)) == -1)
    return false;
  opened = true;

  unsigned char value;
  socklen_t len = sizeof(value);

  value = 1; // force reuse 
  setsockopt(sck, SOL_SOCKET, SO_REUSEADDR, (void *) &value, len );

  if(connect(sck, (struct sockaddr*)&peeraddr_in,
	     sizeof(struct sockaddr_in)) == -1)
    return false;
      
  socklen_t addrlen = sizeof(struct sockaddr_in);
  struct sockaddr_in myaddr_in;
  memset ((char *)&myaddr_in, 0, sizeof(struct sockaddr_in));
      
  if (getsockname(sck, (struct sockaddr*)&myaddr_in, &addrlen) == -1)
    return false;
  
  return InitGSIAuthentication(sck);
}


/**
 * Close the connection.
 * @return true for successful close, false otherwise.
 */
void
GSISocketClient::Close()
{
  OM_uint32 status = 0;

  if (context != GSS_C_NO_CONTEXT)
     gss_delete_sec_context(&status, &context, GSS_C_NO_BUFFER);
  context= GSS_C_NO_CONTEXT;
  if (credential != GSS_C_NO_CREDENTIAL)
    gss_release_cred(&status,&credential);
  credential = GSS_C_NO_CREDENTIAL;
  if (opened)
    close(sck);
  opened=false;
}



/**
 * Send a string value.
 * @param s the string value to send.
 * @return true on success, false otherwise.
 */ 
bool 
GSISocketClient::Send(const std::string s)
{
  bool return_status;

  if (return_status = !(context == GSS_C_NO_CONTEXT)) {

    OM_uint32        maj_stat, min_stat;
     
    int token_status;
    maj_stat = globus_gss_assist_wrap_send (&min_stat,
					    context,
					    const_cast<char *>(s.c_str()),
					    s.length(),
					    &token_status,
					    send_token,
					    &sck,
					    gsi_logfile);

    return_status = !GSS_ERROR(maj_stat);
  }

  return return_status;
}


/**
 * Receive a string value.
 * @param s the string to fill.
 * @return true on success, false otherwise.
 */
bool 
GSISocketClient::Receive(std::string& s)
{
  bool return_status;

  OM_uint32 maj_stat, min_stat;

  char  *message;
  size_t length;
  int    token_status;

  maj_stat = globus_gss_assist_get_unwrap(&min_stat, context,
					  &message, &length, &token_status, 
					  get_token, &sck,
					  gsi_logfile);

  if (return_status = !GSS_ERROR(maj_stat))
    s = std::string(message,length);
  else
    globus_gss_assist_display_status(gsi_logfile, 
				     "GSS authentication failure ",
				     maj_stat, min_stat, token_status); 

  return return_status;
}
